@@ -13,6 +13,7 @@ from typing import (
     Generic,
     Iterable,
     List,
+    Mapping,
     MutableMapping,
     Optional,
     Tuple,
@@ -29,12 +30,18 @@ from di.types.providers import (
     AsyncGeneratorProvider,
     CallableProvider,
     CoroutineProvider,
-    DependencyProvider,
     DependencyType,
     GeneratorProvider,
 )
 
 
+class ExecutionState(typing.NamedTuple):
+    container_state: ContainerState
+    results: Dict[DependantProtocol[Any], Any]
+    dependency_counts: MutableMapping[DependantProtocol[Any], int]
+    dependants: Mapping[DependantProtocol[Any], Iterable[Task[Any]]]
+
+
 class Task(Generic[DependencyType]):
     __slots__ = ("dependant", "dependencies")
 
@@ -48,36 +55,12 @@ class Task(Generic[DependencyType]):
 
     def compute(
         self,
-        state: ContainerState,
-        results: Dict[DependantProtocol[Any], Any],
-        values: typing.Mapping[DependencyProvider, Any],
-        dependency_counts: MutableMapping[DependantProtocol[Any], int],
-        dependants: MutableMapping[DependantProtocol[Any], List[Task[Any]]],
+        state: ExecutionState,
     ) -> Union[
         Awaitable[Iterable[Optional[ExecutorTask]]], Iterable[Optional[ExecutorTask]]
     ]:
         raise NotImplementedError
 
-    def from_cache_or_values(
-        self,
-        state: ContainerState,
-        results: Dict[DependantProtocol[Any], Any],
-        values: typing.Mapping[DependencyProvider, Any],
-    ) -> bool:
-        assert self.dependant.call is not None
-        if self.dependant.call in values:
-            value = values[self.dependant.call]
-            results[self.dependant] = value
-            if self.dependant.share:
-                state.cached_values.set(
-                    self.dependant.call, value, scope=self.dependant.scope
-                )
-            return True
-        elif self.dependant.share and state.cached_values.contains(self.dependant.call):
-            results[self.dependant] = state.cached_values.get(self.dependant.call)
-            return True
-        return False
-
     def gather_params(
         self, results: Dict[DependantProtocol[Any], Any]
     ) -> Tuple[List[Any], Dict[str, Any]]:
@@ -92,37 +75,28 @@ class Task(Generic[DependencyType]):
         return positional, keyword
 
     def gather_new_tasks(
-        self,
-        state: ContainerState,
-        results: Dict[DependantProtocol[Any], Any],
-        values: typing.Mapping[DependencyProvider, Any],
-        dependency_counts: MutableMapping[DependantProtocol[Any], int],
-        dependants: MutableMapping[DependantProtocol[Any], List[Task[Any]]],
+        self, state: ExecutionState
     ) -> Iterable[Optional[ExecutorTask]]:
         """Look amongst our dependants to see if any of them are now dependency free"""
         new_tasks: Deque[Optional[ExecutorTask]] = deque()
-        for dependant in dependants[self.dependant]:
-            count = dependency_counts[dependant.dependant]
+        for dependant in state.dependants[self.dependant]:
+            count = state.dependency_counts[dependant.dependant]
             if count == 1:
                 # this dependant has no further dependencies, so we can compute it now
                 newtask = functools.partial(
                     dependant.compute,
-                    state=state,
-                    results=results,
-                    values=values,
-                    dependency_counts=dependency_counts,
-                    dependants=dependants,
+                    state,
                 )
                 new_tasks.append(newtask)  # type: ignore[arg-type] # because of partial
                 # pop it from dependency counts so that we can
                 # tell when we've satisfied all dependencies below
-                dependency_counts.pop(dependant.dependant)
+                state.dependency_counts.pop(dependant.dependant)
             else:
-                dependency_counts[dependant.dependant] = count - 1
+                state.dependency_counts[dependant.dependant] = count - 1
         # also pop ourselves if we have no deps
-        if dependency_counts.get(self.dependant, -1) == 0:
-            dependency_counts.pop(self.dependant)
-        if not dependency_counts:
+        if state.dependency_counts.get(self.dependant, -1) == 0:
+            state.dependency_counts.pop(self.dependant)
+        if not state.dependency_counts:
             # all dependencies are taken care of, insert sentinel None value
             new_tasks.append(None)
         return new_tasks
@@ -136,18 +110,14 @@ class AsyncTask(Task[DependencyType]):
 
     async def compute(
         self,
-        state: ContainerState,
-        results: Dict[DependantProtocol[Any], Any],
-        values: typing.Mapping[DependencyProvider, Any],
-        dependency_counts: MutableMapping[DependantProtocol[Any], int],
-        dependants: MutableMapping[DependantProtocol[Any], List[Task[Any]]],
+        state: ExecutionState,
     ) -> Iterable[Optional[ExecutorTask]]:
-        args, kwargs = self.gather_params(results)
+        args, kwargs = self.gather_params(state.results)
 
         assert self.dependant.call is not None
         call = self.dependant.call
         if is_async_gen_callable(call):
-            stack = state.stacks[self.dependant.scope]
+            stack = state.container_state.stacks[self.dependant.scope]
             if not isinstance(stack, AsyncExitStack):
                 raise IncompatibleDependencyError(
                     f"The dependency {self.dependant} is an awaitable dependency"
@@ -155,26 +125,15 @@ class AsyncTask(Task[DependencyType]):
                 )
             if TYPE_CHECKING:
                 call = cast(AsyncGeneratorProvider[DependencyType], call)
-            results[self.dependant] = await stack.enter_async_context(
+            state.results[self.dependant] = await stack.enter_async_context(
                 asynccontextmanager(call)(*args, **kwargs)
             )
         else:
             if TYPE_CHECKING:
                 call = cast(CoroutineProvider[DependencyType], call)
-            results[self.dependant] = await call(*args, **kwargs)
-        if self.dependant.share:
-            # caching is allowed, now that we have a value we can save it and start using the cache
-            state.cached_values.set(
-                self.dependant.call, results[self.dependant], scope=self.dependant.scope
-            )
+            state.results[self.dependant] = await call(*args, **kwargs)
 
-        return self.gather_new_tasks(
-            state=state,
-            results=results,
-            values=values,
-            dependency_counts=dependency_counts,
-            dependants=dependants,
-        )
+        return self.gather_new_tasks(state)
 
 
 class SyncTask(Task[DependencyType]):
@@ -182,38 +141,23 @@ class SyncTask(Task[DependencyType]):
 
     def compute(
         self,
-        state: ContainerState,
-        results: Dict[DependantProtocol[Any], Any],
-        values: typing.Mapping[DependencyProvider, Any],
-        dependency_counts: MutableMapping[DependantProtocol[Any], int],
-        dependants: MutableMapping[DependantProtocol[Any], List[Task[Any]]],
+        state: ExecutionState,
     ) -> Iterable[Optional[ExecutorTask]]:
 
-        args, kwargs = self.gather_params(results)
+        args, kwargs = self.gather_params(state.results)
 
         assert self.dependant.call is not None
         call = self.dependant.call
         if is_gen_callable(self.dependant.call):
             if TYPE_CHECKING:
                 call = cast(GeneratorProvider[DependencyType], call)
-            stack = state.stacks[self.dependant.scope]
-            results[self.dependant] = stack.enter_context(
+            stack = state.container_state.stacks[self.dependant.scope]
+            state.results[self.dependant] = stack.enter_context(
                 contextmanager(call)(*args, **kwargs)
             )
         else:
             if TYPE_CHECKING:
                 call = cast(CallableProvider[DependencyType], call)
-            results[self.dependant] = call(*args, **kwargs)
-        if self.dependant.share:
-            # caching is allowed, now that we have a value we can save it and start using the cache
-            state.cached_values.set(
-                self.dependant.call, results[self.dependant], scope=self.dependant.scope
-            )
+            state.results[self.dependant] = call(*args, **kwargs)
 
-        return self.gather_new_tasks(
-            state=state,
-            results=results,
-            values=values,
-            dependency_counts=dependency_counts,
-            dependants=dependants,
-        )
+        return self.gather_new_tasks(state)
