@@ -1,17 +1,22 @@
 from __future__ import annotations
 
 import functools
-from collections import defaultdict, deque
+import typing
+from collections import deque
 from contextvars import ContextVar
+from dataclasses import dataclass
 from typing import (
     Any,
     ContextManager,
-    DefaultDict,
     Deque,
     Dict,
+    FrozenSet,
+    Iterable,
     List,
     Mapping,
+    NamedTuple,
     Optional,
+    Set,
     Tuple,
     Union,
     cast,
@@ -20,8 +25,9 @@ from typing import (
 from di._dag import topsort
 from di._inspect import is_async_gen_callable, is_coroutine_callable
 from di._local_scope_context import LocalScopeContext
+from di._nullcontext import nullcontext
 from di._state import ContainerState
-from di._task import AsyncTask, SyncTask, Task
+from di._task import AsyncTask, ExecutionState, SyncTask, Task
 from di.exceptions import (
     DependencyRegistryError,
     DuplicateScopeError,
@@ -44,6 +50,20 @@ from di.types.solved import SolvedDependency
 Dependency = Any
 
 
+class _ExecutionPlanCache(NamedTuple):
+    cache_key: FrozenSet[DependantProtocol[Any]]
+    dependant_dag: Mapping[DependantProtocol[Any], Iterable[Task[Any]]]
+    dependency_counts: Dict[DependantProtocol[Any], int]
+    leaf_tasks: Iterable[Task[Any]]
+
+
+@dataclass
+class _SolvedDependencyCache:
+    tasks: Mapping[DependantProtocol[Any], Task[Any]]
+    dependency_dag: Mapping[DependantProtocol[Any], Set[DependantProtocol[Any]]]
+    execution_plan: Optional[_ExecutionPlanCache] = None
+
+
 class Container:
     _context: ContextVar[ContainerState]
     _executor: Union[AsyncExecutor, SyncExecutor]
@@ -51,7 +71,7 @@ class Container:
     def __init__(
         self,
         *,
-        default_scope: Scope = None,
+        execution_scope: Scope = None,
         executor: Optional[Union[AsyncExecutor, SyncExecutor]] = None,
     ) -> None:
         self._context = ContextVar("context")
@@ -60,7 +80,7 @@ class Container:
         state.cached_values.set(Container, self, scope="container")
         self._context.set(state)
         self._executor = executor or DefaultExecutor()
-        self._default_scope = default_scope
+        self._execution_scope = execution_scope
 
     @property
     def _state(self) -> ContainerState:
@@ -98,20 +118,12 @@ class Container:
 
         Binds are only identified by the identity of the callable and do not take into account
         the scope or any other data from the dependency they are replacing.
-
-        The `scope` parameter determines the scope for the bind itself.
-        The bind will be automatically cleared when that scope is exited.
-        If no scope is provided, the current scope is used.
         """
         return self._state.bind(provider=provider, dependency=dependency)
 
-    def __contains__(self, o: object) -> bool:
-        return o in self._state.binds
-
-    def __getitem__(
-        self, provider: DependencyProviderType[DependencyType]
-    ) -> DependantProtocol[DependencyType]:
-        return self._state.binds[provider]
+    @property
+    def binds(self) -> Mapping[DependencyProvider, DependantProtocol[Any]]:
+        return self._state.binds
 
     def solve(
         self,
@@ -180,16 +192,13 @@ class Container:
                         q.append(subdep)
         tasks = self._build_tasks(param_graph, topsort(dep_dag))
         dependency_dag = {dep: set(subdeps) for dep, subdeps in dep_dag.items()}
-        # we store several things:
-        # - the dependency itself
-        # - dag, which includes parameter information and is public
-        # - _dependency_dag: same as dag but without param info and w/o repeated params (so a set instead of a list)
-        # - _tasks: a mapping of dependency to tasks
+        container_cache = _SolvedDependencyCache(
+            dependency_dag=dependency_dag, tasks=tasks
+        )
         return SolvedDependency(
             dependency=dependency,
             dag=param_graph,
-            _dependency_dag=dependency_dag,
-            _tasks=tasks,
+            container_cache=container_cache,
         )
 
     def _build_tasks(
@@ -267,58 +276,122 @@ class Container:
         *,
         validate_scopes: bool = True,
         values: Optional[Mapping[DependencyProvider, Any]] = None,
-    ) -> Tuple[Dict[DependantProtocol[Any], Any], List[ExecutorTask]]:
-        results: Dict[DependantProtocol[Any], Any] = {}
-        values = values or {}
+    ) -> Tuple[
+        Dict[DependantProtocol[Any], Any],
+        List[ExecutorTask],
+        typing.List[DependantProtocol[Any]],
+    ]:
+        user_values = values or {}
         if validate_scopes:
             self._validate_scopes(solved)
 
-        tasks = solved._tasks
+        if type(solved.container_cache) is not _SolvedDependencyCache:
+            raise TypeError(
+                "This SolvedDependency was not created by this Container"
+            )  # pragma: no cover
 
-        # Build a DAG of Tasks that we actually need to execute
-        # this allows us to prune subtrees that will come from cached values
-        unvisited = deque([solved.dependency])
-        # Make DAG values a set to account for dependencies that depend on the the same
-        # sub dependency in more than one param (`def func(a: A, a_again: A)`)
-        dag: Dict[DependantProtocol[Any], List[DependantProtocol[Any]]] = {}
-        dependant_dag: DefaultDict[
-            DependantProtocol[Any], List[Task[Any]]
-        ] = defaultdict(list)
-        while unvisited:
-            dep = unvisited.pop()
-            if dep in dag:
-                continue
-            task = tasks[dep]
-            # task the dependency is cached or was provided by value
-            # we don't need to compute it or any of it's dependencies
-            if not task.from_cache_or_values(self._state, results, values):
-                # otherwise, we add it to our DAG and visit it's children
-                dag[dep] = []
-                for subdep in solved._dependency_dag[dep]:
-                    if not tasks[subdep].from_cache_or_values(
-                        self._state, results, values
-                    ):
-                        dependant_dag[subdep].append(tasks[dep])
-                        dag[dep].append(subdep)
-                        if subdep not in dag:
-                            unvisited.append(subdep)
+        solved_dependency_cache = solved.container_cache
 
-        dependency_counts = {dep: len(subdeps) for dep, subdeps in dag.items()}
+        cache: Dict[DependencyProvider, Any] = {}
+        for mapping in self._state.cached_values.mappings.values():
+            cache.update(mapping)
+        cache.update(user_values)
 
-        leafs: List[ExecutorTask] = []
-        for dep, depcount in dependency_counts.items():
-            if depcount == 0:
-                leaf = functools.partial(
-                    tasks[dep].compute,
-                    state=self._state,
-                    results=results,
-                    values=values,
-                    dependency_counts=dependency_counts,
-                    dependants=dependant_dag,
-                )
-                leafs.append(leaf)  # type: ignore[arg-type] # beacuse of partial
+        to_cache: typing.List[DependantProtocol[Any]] = []
+        execution_scope = self.scopes[-1]
 
-        return results, leafs
+        results: Dict[DependantProtocol[Any], Any] = {}
+
+        def use_cache(dep: DependantProtocol[Any]) -> None:
+            call = dep.call
+            if call in cache:
+                if call in user_values:
+                    results[dep] = user_values[call]
+                    return
+                elif dep.share:
+                    results[dep] = cache[call]
+                    return
+            else:
+                if dep.share and dep.scope != execution_scope:
+                    to_cache.append(dep)
+            return
+
+        for dep in solved.dag:
+            use_cache(dep)
+
+        execution_plan_cache_key = frozenset(results.keys())
+
+        execution_plan = solved.container_cache.execution_plan
+
+        if (
+            execution_plan is None
+            or execution_plan.cache_key != execution_plan_cache_key
+        ):
+            # Build a DAG of Tasks that we actually need to execute
+            # this allows us to prune subtrees that will come
+            # from pre-computed values (values paramter or cache)
+            unvisited = deque([solved.dependency])
+            # Make DAG values a set to account for dependencies that depend on the the same
+            # sub dependency in more than one param (`def func(a: A, a_again: A)`)
+            dependency_counts: Dict[DependantProtocol[Any], int] = {}
+            dependant_dag: Dict[DependantProtocol[Any], Deque[Task[Any]]] = {
+                dep: deque() for dep in solved_dependency_cache.dependency_dag
+            }
+            while unvisited:
+                dep = unvisited.pop()
+                if dep in dependency_counts:
+                    continue
+                # task the dependency is cached or was provided by value
+                # we don't need to compute it or any of it's dependencies
+                if dep not in results:
+                    # otherwise, we add it to our DAG and visit it's children
+                    dependency_counts[dep] = 0
+                    for subdep in solved_dependency_cache.dependency_dag[dep]:
+                        if subdep not in results:
+                            dependant_dag[subdep].append(
+                                solved_dependency_cache.tasks[dep]
+                            )
+                            dependency_counts[dep] += 1
+                            if subdep not in dependency_counts:
+                                unvisited.append(subdep)
+
+            solved.container_cache.execution_plan = (
+                execution_plan
+            ) = _ExecutionPlanCache(
+                cache_key=execution_plan_cache_key,
+                dependant_dag=dependant_dag,
+                dependency_counts=dependency_counts,
+                leaf_tasks=[
+                    solved_dependency_cache.tasks[dep]
+                    for dep, count in dependency_counts.items()
+                    if count == 0
+                ],
+            )
+
+        state = ExecutionState(
+            container_state=self._state,
+            results=results,
+            dependency_counts=execution_plan.dependency_counts.copy(),
+            dependants=execution_plan.dependant_dag,
+        )
+
+        return (
+            results,
+            [functools.partial(t.compute, state) for t in execution_plan.leaf_tasks],
+            to_cache,
+        )
+
+    def _update_cache(
+        self,
+        results: Dict[DependantProtocol[Any], Any],
+        to_cache: Iterable[DependantProtocol[Any]],
+    ) -> None:
+        for dep in to_cache:
+            self._state.cached_values.set(
+                dep.call,  # type: ignore[arg-type]
+                results[dep],
+                scope=dep.scope,
+            )
 
     def execute_sync(
         self,
@@ -332,8 +405,13 @@ class Container:
         If you are not dynamically changing scopes, you can run once with `validate_scopes=True`
         and then disable scope validation in subsequent runs with `validate_scope=False`.
         """
-        with self.enter_local_scope(self._default_scope):
-            results, queue = self._prepare_execution(
+        cm: FusedContextManager[None]
+        if self._execution_scope in self.scopes:
+            cm = nullcontext(None)
+        else:
+            cm = self.enter_local_scope(self._execution_scope)
+        with cm:
+            results, queue, to_cache = self._prepare_execution(
                 solved, validate_scopes=validate_scopes, values=values
             )
             if not hasattr(self._executor, "execute_sync"):
@@ -343,7 +421,10 @@ class Container:
             executor = cast(SyncExecutor, self._executor)
             if queue:
                 executor.execute_sync(queue)
-            return results[solved.dependency]
+
+            res = results[solved.dependency]
+            self._update_cache(results, to_cache)
+            return res
 
     async def execute_async(
         self,
@@ -357,8 +438,13 @@ class Container:
         If you are not dynamically changing scopes, you can run once with `validate_scopes=True`
         and then disable scope validation in subsequent runs with `validate_scope=False`.
         """
-        async with self.enter_local_scope(self._default_scope):
-            results, queue = self._prepare_execution(
+        cm: FusedContextManager[None]
+        if self._execution_scope in self.scopes:
+            cm = nullcontext(None)
+        else:
+            cm = self.enter_local_scope(self._execution_scope)
+        async with cm:
+            results, queue, to_cache = self._prepare_execution(
                 solved, validate_scopes=validate_scopes, values=values
             )
             if not hasattr(self._executor, "execute_async"):
@@ -368,4 +454,6 @@ class Container:
             executor = cast(AsyncExecutor, self._executor)
             if queue:
                 await executor.execute_async(queue)
-            return results[solved.dependency]
+            res = results[solved.dependency]
+            self._update_cache(results, to_cache)
+            return res
